/*
 * caps.h - Classes for Entity Capabilities (JEP-115)
 * Copyright (C) 2005  Remko Troncon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef CAPS_H
#define CAPS_H

#include <qobject.h>
#include <qtimer.h>
#include <qpair.h>
#include <qdatetime.h>

#include "im.h"
#include "psiaccount.h"

class CapsManager : public QObject
{
	Q_OBJECT

public:
	static CapsManager* instance();

	bool isEnabled();
	void setEnabled(bool);
	void reset();
	void setFile(const QString&);
	void removeAccount(PsiAccount*);

	void updateCaps(PsiAccount* acc, const Jid& jid, const QString& node, const QString& ver, const QString& ext);
	void disableCaps(const Jid& jid);
	bool capsEnabled(const Jid& jid) const;
	XMPP::Features features(const Jid& jid) const;
	QString clientName(const Jid& jid) const;
	QString clientVersion(const Jid& jid) const;
	
signals:
	/**
	 * This signal is emitted when the feature list of a given JID have changed.
	 */
	void capsChanged(const Jid& jid);

protected slots:
	void discoFinished();

protected:
	void requestDiscoInfo(PsiAccount* acc, const Jid& jid, const QString& node);
	void save() const;
	
public:
	class CapsSpec;
	typedef QValueList<CapsSpec> CapsSpecs;

	class CapsSpec
	{
		public:
			CapsSpec();
			CapsSpec(const QString&, const QString&, const QString&);
			const QString& node() const; 
			const QString& version() const; 
			const QString& extensions() const; 
			CapsSpecs flatten() const;

			bool operator==(const CapsSpec&) const;
			bool operator!=(const CapsSpec&) const;
			bool operator<(const CapsSpec&) const;
			
		private:
			QString node_, ver_, ext_;
	};

	class CapsInfo
	{
		public:
			CapsInfo();
			const QStringList& features() const;
			const DiscoItem::Identities& identities() const;
			QStringList jids() const;
			bool discovered() const;
			int pendingRequests() const;

			void reset();
			void removeAccount(PsiAccount* acc);
			void removeJid(const Jid&);
			void addJid(const Jid&, PsiAccount*);
			QPair<Jid,PsiAccount*> nextJid(const Jid&, const Task*);
			
			void setDiscovered(bool);
			void setPendingRequests(int);
			void setIdentities(const DiscoItem::Identities&);
			void setFeatures(const QStringList&);
			
			QDomElement toXml(QDomDocument *) const;
			void fromXml(const QDomElement&);

		protected:
			void updateLastSeen();
			
		private:
			bool discovered_;
			int pendingRequests_;
			QStringList features_;
			DiscoItem::Identities identities_;
			QValueList<QPair<QString,PsiAccount*> > jids_;
			QDate lastSeen_;
	};

	CapsManager();
	static CapsManager* instance_;
	
	bool isEnabled_;
	QString fileName_;
	QMap<QString,CapsSpec> capsSpecs_;
	QMap<CapsSpec,CapsInfo> capsInfo_;
};


#endif
